package de.hsbo.fbv.bmg.graphics.transforms;
import java.awt.Shape;
import java.awt.geom.Path2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;

/**
 * Allgemeine (abstrakte) Klasse fr Transformationen im R
 * Bekannte Kindklasse: Bilinear Transform
 * @author A.Zimmermann, HS Bochum, 2011
 *
 */
public abstract class Transform2D {

	/**
	 * Transformiert eine geometrische Figur vom Typ Shape.
	 * @param s - Shape im Urbild
	 * @return Shape im Abbild
	 */
	public Shape transform(Shape s) {
		PathIterator pi = s.getPathIterator(null);
		Path2D p = new Path2D.Double();
		double[] src = new double[6];
		double[] dst = new double[6];
		while(!pi.isDone()) {
			int type = pi.currentSegment(src);
			dst = transform(src);
			if(type==PathIterator.SEG_MOVETO) {
				p.moveTo(dst[0], dst[1]);
			}
			else if(type==PathIterator.SEG_LINETO) {
				p.lineTo(dst[0], dst[1]);
			}
			else if(type==PathIterator.SEG_QUADTO) {
				p.quadTo(dst[0], dst[1], dst[2], dst[3]);
			}
			else if(type==PathIterator.SEG_CUBICTO) {
				p.curveTo(dst[0], dst[1], dst[2], dst[3], dst[4], dst[5]);
			}
			else if(type==PathIterator.SEG_CLOSE) {
				p.closePath();
			}
			else
				;
			pi.next();
		}
		return p;
	}

	/**
	 * Transformiert eine geometrische Figur vom Typ Shape ins Urbild zurck.
	 * @param s - Shape im Abbild
	 * @return Shape im Urbild
	 */
	public Shape inverseTransform(Shape s) {
		PathIterator pi = s.getPathIterator(null);
		Path2D p = new Path2D.Double();
		double[] src = new double[6];
		double[] dst = new double[6];
		while(!pi.isDone()) {
			int type = pi.currentSegment(src);
			dst = transform(src);
			if(type==PathIterator.SEG_MOVETO) {
				p.moveTo(dst[0], dst[1]);
			}
			else if(type==PathIterator.SEG_LINETO) {
				p.lineTo(dst[0], dst[1]);
			}
			else if(type==PathIterator.SEG_QUADTO) {
				p.quadTo(dst[0], dst[1], dst[2], dst[3]);
			}
			else if(type==PathIterator.SEG_CUBICTO) {
				p.curveTo(dst[0], dst[1], dst[2], dst[3], dst[4], dst[5]);
			}
			else if(type==PathIterator.SEG_CLOSE) {
				p.closePath();
			}
			else
				;
			pi.next();
		}
		return p;
	}
	
	/**
	 * Transformiert einen Punkt. 
	 * @param point - Punkt im Urbild
	 * @return - Punkt im Abbild
	 */
	public Point2D transform(Point2D point) {
		double[] coords = transform(point.getX(), point.getY());
		return new Point2D.Double(coords[0], coords[1]);
	}
	
	/**
	 * Transformiert einen Punkt ins Urbild zurck.
	 * @param point - Punkt im Abbild
	 * @return - Punkt im Urbild
	 */
	public Point2D inverseTransform(Point2D point) {
		double[] coords = inverseTransform(point.getX(), point.getY());
		return new Point2D.Double(coords[0], coords[1]);
	}

	/**
	 * Transformiert ein Koordinaten-Array {x0, y0, x1, y1, ...}.
	 * Die Methode wird intern zur Transformation von Shapes eingesetzt.
	 * @param srcPts - Koordinatenfeld im Urbild
	 * @return - Koordinatenfeld im Abbild
	 */
	public double[] transform(double[] srcPts) {
	double[] dstPts = new double[srcPts.length];
	int srcOff = 0;
	int dstOff = 0; 
	int numPts=srcPts.length/2;
	while(numPts>0) {
		double x = srcPts[srcOff++];
		double y = srcPts[srcOff++];
		double[] result = transform(x, y);
		dstPts[dstOff++] = result[0];
		dstPts[dstOff++] = result[1];
		numPts--;
	}
	return dstPts;
	}
	
	/**
	 * Transformiert ein Koordinaten-Array {x0, y0, x1, y1, ...} ins Urbild zurck.
	 * Die Methode wird intern zur Rcktransformation von Shapes eingesetzt.
	 * @param srcPts - Koordinatenfeld im Abbild
	 * @return - Koordinatenfeld im Urbild
	 */
	public double[] inverseTransform(double[] srcPts) {
		double[] dstPts = new double[srcPts.length];
		int srcOff = 0;
		int dstOff = 0; 
		int numPts=srcPts.length/2;
		while(numPts>0) {
			double x = srcPts[srcOff++];
			double y = srcPts[srcOff++];
			double[] result = inverseTransform(x, y);
			dstPts[dstOff++] = result[0];
			dstPts[dstOff++] = result[1];
			numPts--;
		}
		return dstPts;
		}
		
	/**
	 * Abstrakte Methode: Transformiert ein Koordinatenpaar.
	 * Die Methode wird intern von anderen Methoden dieser Klasse verwendet.
	 * @param x - x-Koordinate im Urbild
	 * @param y - y-Koordinate im Urbild 
	 * @return - Koordinatenpaar im Abbild
	 */
	public abstract double[] transform(double x, double y);
	
	/**
	 * Abstrakte Methode: Transformiert ein Koordinatenpaar ins Urbild zurck.
	 * Die Methode wird intern von anderen Methoden dieser Klasse verwendet.
	 * @param x - x-Koordinate im Abbild
	 * @param y - y-Koordinate im Abbild 
	 * @return - Koordinatenpaar im Urbild
	 */
	public abstract double[] inverseTransform(double x, double y);
	
}

