package de.hsbo.fbv.bmg.graphics.transforms;

import Jama.Matrix;

/**
 * Klasse zur Abbildung mit einer dreidimensionalen Transformation der Form:
 * [x']   [m00 m01 m02 m03] [x]   [m00*x + m01*y + m02*z + m03]
 * [y'] = [m10 m11 m12 m13]*[y] = [m10*x + m11*y + m12*z + m13]
 * [z']   [m20 m21 m22 m23] [z]   [m20*x + m21*y + m22*z + m23]
 * [1 ]   [ 0   0   0   1 ] [1]   [                         1 ]
 * 
 * 
 * @author A.Zimmermann, HS Bochum, 2011
 *
 */
public class Transform2hD {
	
	private static double eps = 1.e-15;
	public static int PROJ_NONE = 0;
	public static int PROJ_PARALLEL = 1;
	public static int PROJ_CENTRAL = 2;
	public static int PROJ_OPTICAL = 3;
	
	double[][] a = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
	Transform2hD inverse = null;
	int projection = 0;
	boolean changed = false;
	
	
	/**
	 * Standardkonstruktor
	 */
	public Transform2hD() {
		super();
	}
	
	/**
	 * Erzeugt eine 2hD-Transformation mit dem Array einer 4x4-Matrix.
	 * @param a
	 */
	private Transform2hD(double[][] a) {
		this.a = a;
		this.changed=true;
	}
	
	/**
	 * Fgt dieser Transformation (this) die Teiltransformation trans hinzu.
	 * Anders als bei AffineTransform sind aber die Teiltransformationen 
	 * in der richtigen Reihenfolge anzugeben, da die entsprechenden Matrizen 
	 * von Links multipliziert werden. Die Methode wird intern fr verwendet.
	 * @param trans
	 */
	public void concatenate(Transform2hD trans) {
		Matrix mtrans = new Matrix(trans.a);
		Matrix mthis = new Matrix(this.a);
		Matrix result = mtrans.times(mthis);
		this.a = result.getArray();
		this.changed=true;
	}

	/**
	 * Fgt dieser Transformation eine Translation hinzu.
	 * @param tx
	 * @param ty
	 * @param tz
	 */
	public void translate(double tx, double ty, double tz) {
		Transform2hD trans = new Transform2hD(new double[][] {
				{1, 0, 0, tx},
				{0, 1, 0, ty},
				{0, 0, 1, tz},
				{0, 0, 0,  1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation eine Rotation um x hinzu.
	 * @param rx Drehwinkel in [rad]
	 */
	public void rotateX(double rx) {
		Transform2hD trans = new Transform2hD(new double[][] {
				{1, 0,             0,            0},
				{0, Math.cos(rx), -Math.sin(rx), 0},
				{0, Math.sin(rx),  Math.cos(rx), 0},
				{0,             0,            0, 1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation eine Rotation um y hinzu.
	 * @param ry Drehwinkel in [rad]
	 */
	public void rotateY(double ry) {
		Transform2hD trans = new Transform2hD(new double[][] {
				{ Math.cos(ry), 0, Math.sin(ry), 0},
				{ 0,            1,            0, 0},
				{-Math.sin(ry), 0, Math.cos(ry), 0},
				{0,             0,            0, 1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation eine Rotation um z hinzu.
	 * @param rz Drehwinkel in [rad]
	 */
	public void rotateZ(double rz) {
		Transform2hD trans = new Transform2hD(new double[][] {
				{Math.cos(rz), -Math.sin(rz), 0, 0},
				{Math.sin(rz),  Math.cos(rz), 0, 0},
				{0,             0,            1, 0},
				{0,             0,            0, 1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation Skalierungen der Achsen hinzu.
	 * @param mx Mastabsfaktor der x-Achse
	 * @param mx Mastabsfaktor der y-Achse
	 * @param mx Mastabsfaktor der z-Achse
	 */
	public void scale(double mx, double my, double mz) {
		Transform2hD trans = new Transform2hD(new double[][] {
				{mx,  0,  0, 0},
				{ 0, my,  0, 0},
				{ 0,  0, mz, 0},
				{ 0,  0,  0, 1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation eine Projektion hinzu. Die Projektion wird immer
	 * als letzte Teiltransformation ausgefhrt.
	 * @param projection PROJ_CENTRAL oder PROJ_PARALLEL
	 */
	public void addProjection(int projection) {
		if(projection == 1)
			this.projection = PROJ_PARALLEL;
		else if(projection == 2)
			this.projection = PROJ_CENTRAL;
		else if(projection == 3)
			this.projection = PROJ_OPTICAL;
		else 
			this.projection = 0;
	}
	
	/**
	 * Zieht die gesetzte Projektion zurck.
	 */
	public void removeProjection() {
		this.projection = 0;
	}
	
	/**
	 * Transformiert ein Koordinaten-Tripel.
	 * @param x
	 * @param y
	 * @param z
	 * @return transformierte Koordinaten als Array
	 */
	public double[] transform(double x, double y, double z) {
		// 3D-Transformation durchfhren
		double xd = (a[0][0] * x + a[0][1] * y + a[0][2] * z + a[0][3]);
		double yd = (a[1][0] * x + a[1][1] * y + a[1][2] * z + a[1][3]);
		double zd = (a[2][0] * x + a[2][1] * y + a[2][2] * z + a[2][3]);
		// keine Projektion
		if(projection==0) {
			return new double[] {xd, yd, zd};
		}
		// Parallelprojektion
		else if(projection==PROJ_PARALLEL) {
			if(zd<-1)
				return new double[] {xd, yd};
			else return null;
		}
		// Zentralprojektion
		else if(projection==PROJ_CENTRAL) {
			if(zd<-1)
				return new double[] {xd/zd, yd/zd};
			else return null;
		}
		// Zentralprojektion mit Umkehrung des Bildes
		else if(projection==PROJ_OPTICAL) {
			if(zd<-1)
				return new double[] {-xd/zd, -yd/zd};
			else return null;
		}
		else
			return null;
	}
		
	/**
	 * Berechnet die inverse Transformation. Das Ergebnis reprsentiert einen 
	 * Richtungsvektor, der in den Objektraum zeigt.
	 * @param x
	 * @param y
	 * @return
	 */
	public double[] inverseTransform(double x, double y) {
		return inverseTransform(x,y,1.0);
	}
	
	/**
	 * Berechnet die inverse Transformation. 
	 * @param x
	 * @param y
	 * @param z
	 * @return
	 */
	public double[] inverseTransform(double x, double y, double z) {
		// inverse Matrix der Transformation beschaffen
		if(this.changed==true) {
			inverseMatrix();			
		}
		return inverse.transform(x, y, z);
	}
	
	
	private void inverseMatrix() {
		Matrix A = new Matrix(this.a);
		Matrix INV = A.inverse();
		this.inverse = new Transform2hD(INV.getArray());
		this.changed = false;
	}
	
	@Override
	public String toString() {
		String s = "";
		for(int i=0; i<this.a.length; i++)
			s = s+"["+rnd(a[i][0])+" "+rnd(a[i][1])+" "+rnd(a[i][2])+" "+rnd(a[i][3])+"]\n";
		return s;
	}
	
	private double rnd(double x) {
		return Math.rint(x/eps)*eps;
	}
	

}
