package de.hsbo.fbv.bmg.graphics.viewer.simple;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JPanel;

/**
 * Map - Klasse zur Darstellung von Karteninhalten in einem Anzeigefenster. Die Inhalt muss nach 
 * Anzeigeebenen (Folien, Layer) gegliedert sein. Die Klasse untersttzt keine interaktive Manipulation 
 * des Karteninhalts.<br>  
 * Map ist von JPanel abgeleitet und kann daher in andere Anwnedungsfenster eingebunden werden.<br>
 * @author Zimmermann, HS Bochum 2011<br>
 * @see JPanel
 * @see Layer
 */
public class Map extends JPanel {
	
	private static final long serialVersionUID = 1L;
	
	// Konstanten und Klassenattribute
	public static final Color DRAWCOLOR = Color.BLACK;	
	public static final Color FILLCOLOR = Color.GRAY;
	public static final Stroke STROKE = new BasicStroke();
	public static final Paint PAINT = FILLCOLOR;
	public static double factor = 0.96;
	
	// Attribute
	List<Layer> list = new ArrayList<Layer>();
	Rectangle2D window = null;

	public int numLayers() {
		return list.size();
	}

	public boolean isEmpty() {
		return list.isEmpty();
	}

	public boolean contains(Layer o) {
		return list.contains(o);
	}

	public boolean add(Layer e) {
		boolean result = list.add(e);
		this.repaint();
		return result;
	}

	public boolean remove(Layer o) {
		return list.remove(o);
	}

	public void clear() {
		list.clear();
	}

	public Layer get(int index) {
		return list.get(index);
	}

	public Layer set(int index, Layer element) {
		Layer result = list.set(index, element);
		this.repaint();
		return result;
	}

	public void add(int index, Layer element) {
		list.add(index, element);
		this.repaint();
	}

	public Layer removeLayer(int index) {
		Layer result = list.remove(index);
		this.repaint();
		return result;
	}

	public int indexOf(Layer o) {
		return list.indexOf(o);
	}

	public int lastIndexOf(Layer o) {
		return list.lastIndexOf(o);
	}

	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;
		g2.setTransform(this.getTransform());
		for (int i = 0; i < list.size(); i++) {
			if (!list.get(i).isVisible())
				;
			else {
				StyledShapes[] stshapes = list.get(i).getStyledShapes();
				if(stshapes==null)
					return;
				for (int j = 0; j < stshapes.length; j++) {
					if (stshapes[j].isDraw()) {
						Color color = stshapes[j].getColor();
						Stroke stroke = stshapes[j].getStroke();
						if (color == null)
							color = DRAWCOLOR;
						if (stroke == null)
							stroke = STROKE;
						g2.setColor(color);
						g2.setStroke(stroke);
						Shape[] shapes = stshapes[j].getShapes();
						for (int k = 0; k < shapes.length; k++)
							g2.draw(shapes[k]);
					} else if (stshapes[j].isFill()) {
						Color color = stshapes[j].getColor();
						if (color == null)
							color = FILLCOLOR;
						g2.setColor(color);
						Shape[] shapes = stshapes[j].getShapes();
						for (int k = 0; k < shapes.length; k++)
							g2.fill(shapes[k]);
					} else ;
				}
			}
		}
	} // Ende paint()
	
	protected AffineTransform getTransform() {
		if(this.list.isEmpty())
			return new AffineTransform();
		else if (this.window==null)
			return getTransform(this.getBounds2D(), this.getBounds());
		else
			return getTransform(this.window, this.getBounds());
	}

	protected AffineTransform getTransform(Rectangle2D world, Rectangle device) {
		// neues AffinTransform-Objekt als Einheitsmatrix
		AffineTransform at = new AffineTransform();
		// Vorbereitung der Transformationsparameter:
		// 1. Verschiebung des Mittelpunktes von bounds in den
		// Koordinatenursprung
		double centerWX = -(world.getCenterX());
		double centerWY = -(world.getCenterY());
		// 2. Skalierung der Achsen x und y, sowie Spiegelung der y-Achse
		// Der Faktor (96%) bewirkt, dass auch alle Shapes im Panel sichtbar
		// sind.
		double scaleX = device.getWidth() * factor / world.getWidth();
		double scaleY = device.getHeight() * factor / world.getHeight();
		double scale = Math.min(scaleX, scaleY);
		scaleX = scale;
		scaleY = -scale;
		// 3. (Rck-) Verschiebung in den Mittelpunkt des Panels
		double centerDX = device.getCenterX();
		double centerDY = device.getCenterY();
		// Durchfhrung der Teiltransfomationen in umgekehrter Reihenfolge:
		// 3. (Rck-) Verschiebung in den Mittelpunkt des Panels
		at.translate(centerDX, centerDY);
		// 2. Skalierung der Achsen x und y, sowie Spiegelung der y-Achse
		at.scale(scaleX, scaleY);
		// 1. Verschiebung des Mittelpunktes von bounds in den
		// Koordinatenursprung
		at.translate(centerWX, centerWY);
		// Ende: Rckgabe des Objektes
		return at;
	}
	
	protected Rectangle2D getBounds2D() {
		Path2D path = new Path2D.Double();
		Rectangle2D bounds = null;
		for(int i=0; i<list.size(); i++) {
			Shape bnd = list.get(i).getBounds2D();
			if(bnd!=null)
				path.append(bnd, false);
		}
		bounds = path.getBounds2D();
		return bounds;
	}

}
