package de.hsbo.fbv.bmg.graphics.viewer.simple;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Rectangle2D;
import javax.swing.JPanel;

/**
 * Simple Shape Viewer - Einfacher Betrachter fr Java2D-Shapes<br> 
 * Dieser Betrachter ist dazu bestimmt, einfache Java2D-Vektor-Grafiken anzuzeigen,
 * die die Implementierungvorgaben der Schnittstelle java.awt.Shape erfllen.
 * Er bietet keine interaktiven Steuerungen (Zoom oder Pan) und kann Graphiken 
 * nur in einer einheitlichen Farbe zeichnen bzw. fllen. Allerdings wird die Grafik an 
 * die Gre des Anwendungsfensters (Frame) angepasst.<br>
 * Die Klasse ist von javax.swing.JPanel abgeleitet und lsst sich in andere Frames 
 * einbauen. Als Beispielklasse liegt SimpleShapeFrame vor. <br>
 * Anmerkung:<br>
 * Das Grundgerst dieser Klasse wurde mit Eclipse Visual Editor (jve) generiert. Die 
 * Kommentare zu den generierten Teilen bleiben unverndert. <br>
 * @author Zimmermann, HS Bochum 2009
 * @see SimpleShapeFrame
 * @version 2011-03-31: Linienmodus (Stroke) festlegen
 */
public class SimpleShapeViewer extends JPanel {

	// Attribute
	private static final long serialVersionUID = 1L;
	
	Path2D path = null;					// Path2D-Objekt, das alle Shapes aufnimmt
	Color drawColor = Color.BLACK;		// Linienfarbe, hier schwarz  //  @jve:decl-index=0:
	Color fillColor = null;				// Fllfarbe, hier nicht belegt.
	Stroke stroke = null;				// Linienmodus (Stroke - Basicstroke)
	static double factor = 0.96;		// alle Grafiken werden auf 96% skaliert, 
										// um sie vollstndig im Fenster anzuzeigen.

	/**
	 * This is the default constructor
	 */
	public SimpleShapeViewer() {
		super();
		this.path = new Path2D.Double();	// Objekt zu path
		initialize();
	}

	/**
	 * This method initializes this
	 * 
	 * @return void
	 */
	private void initialize() {
		this.setSize(280, 150);				// dies ist nur die anfngliche Ausdehnung
		this.setLayout(null);
	}
	
	/**
	 * Kernmethode zur Ausgabe der Shape-Objekte
	 * Die Methode paint() wird intern (z.B. von repaint()) zur Ausgabe des JPanels aufgerufen.
	 * Zuerst werden die Shape-Objekte gefllt, und dann erst ihre Begrenzungslinien
	 * gezeichnet. Voraussetzung ist, dass eine Linien- bzw. eine Fllfarbe gesetzt wurde.
	 * 
	 * @param Graphics g - Grafik-Kontext (Handle) fr das JPanel-Objekt
	 * @return void
	 */
	public void paint(Graphics g) {

		Graphics2D g2 = (Graphics2D)g;			// Java2D-Graphikkontext holen
		g2.setTransform(this.getTransform());	// Tranformation berechnen und g2 zuweisen
		if(this.fillColor!=null) {				// zuerst das Fllen der Grafik behandeln
			g2.setColor(this.fillColor);		// Fllfarbe zuweisen, wenn icht mit null belegt
			g2.fill(this.path);					// alle Shapes (als Ganzes) ausfllen
		}
		if(this.drawColor!=null) {				// dann erst die Begrenzungslinien zeichnen
			g2.setColor(this.drawColor);		// Linienfarbe setzen, wenn nicht mit null belegt
			if(this.stroke!=null)				// Linienmodus setzen, falls angegeben
				g2.setStroke(stroke);
			g2.draw(this.path);					// alle Shapes (als Ganzes) zeichnen
		}
	}

	/**
	 * Transformation vom Weltkoordinatensystem (WCS) ins Gertekoordinatensystem (DCS)
	 * Die Methode fhrt eine verkettete Transformation in drei Schritten durch. Zur Berechnung 
	 * greift sie auf die Ausdehnung von this.path und JPanel (Elternklasse) zu. Sie wird in der
	 * Methode paint() zur Skalierung der Shapes eingesetzt.
	 * Anmerkung:
	 * Die Klasse AffineTransform ist intern als 2x3-Matrix (Homogene Koordinaten) organisiert, 
	 * fr die Regel der Rechtsmultiplikation festgelegt ist. Dies fhrt dazu, dass am Schluss die 
	 * Teiltransformationen in umgekehrter Reihenfolge vollzogen werden mssen.
	 * 
	 * @return AffineTransform - Handle der Transformationsparameter
	 */
	protected AffineTransform getTransform() {
		// neues AffinTransform-Objekt als Einheitsmatrix 
		AffineTransform at = new AffineTransform();
		// Normalfall
		try {
			// Umschreibendes Rechteck bounds aller Shapes bilden
			Rectangle2D bounds = this.path.getBounds2D();
			// Vorbereitung der Transformationsparameter:
			// 1. Verschiebung des Mittelpunktes von bounds in den Koordinatenursprung
			double centerWX = -(bounds.getCenterX());
			double centerWY = -(bounds.getCenterY());
			// 2. Skalierung der Achsen x und y, sowie Spiegelung der y-Achse 
			//    Der Faktor (96%) bewirkt, dass auch alle Shapes im Panel sichtbar sind.
			Rectangle r = this.getBounds();
			double scaleX = r.getWidth()*factor / bounds.getWidth();
			double scaleY = r.getHeight()*factor / bounds.getHeight();
			double scale = Math.min(scaleX, scaleY);
			scaleX = scale;
			scaleY = -scale;
			// 3. (Rck-) Verschiebung in den Mittelpunkt des Panels
			double centerDX = r.getCenterX();
			double centerDY = r.getCenterY();
			// Durchfhrung der Teiltransfomationen in umgekehrter Reihenfolge:
			// 3. (Rck-) Verschiebung in den Mittelpunkt des Panels
			at.translate(centerDX, centerDY);
			// 2. Skalierung der Achsen x und y, sowie Spiegelung der y-Achse 
			at.scale(scaleX, scaleY);
			// 1. Verschiebung des Mittelpunktes von bounds in den Koordinatenursprung
			at.translate(centerWX, centerWY);
			// Ende: Rckgabe des Objektes
			return at; 
		} 
		// Ausnahme: falls noch keine Shapes existieren, muss die Einheitsmatrix ausgegeben werden		
		catch (Exception e) {
			return at;
		}
	}

	/**
	 * Fgt ein weiteres Shape-Objekt hinzu.
	 * Die Methode bernimmt das Shape in das Path2D-Objekt, in dem alle Grafiken gesammelt und
	 * zur Ausgabe bereitgestellt werden. Anschlieend veranlasst repaint() die Aktualisierung der 
	 * Anzeige des Panels.
	 * 
	 * @param shape - Grafik-Objekt
	 * @return void
	 */
	public void add(Shape shape) {
		this.path.append(shape, false);
		this.repaint();
	}

	/**
	 * Setzt die Farbe der Begrenzungslinien von Graphikobjekten.
	 * Mit dem Parameter drawColor=null wird die Anzeige der Begrenzungslininen verhindert.
	 * @param drawColor - Frabe der Begrenbzungslinien
	 * @return void
	 */
	public void setDrawColor(Color drawColor) {
		this.drawColor = drawColor;
	}

	/**
	 * Setzt die Farbe zum Aufllen der Graphikobjekte.
	 * Mit dem Parameter fillColor=null wird die Anzeige der Fllflchen verhindert.
	 * @param fillColor - Frabe der Begrenbzungslinien
	 * @return void
	 */
	public void setFillColor(Color fillColor) {
		this.fillColor = fillColor;
	}

	/**
	 * Setzt den Linienmodus
	 * @param stroke
	 */
	public void setStroke(Stroke stroke) {
		this.stroke = stroke;
	}
	
	
}
